import re, string, sys, math, os
import OBGrep

class Matcher:
    dbCount        = 0    # keeps a count on the number of databases loaded
    dbEntries      = {}   # stores the number of database entries for each DB
    dbPlain        = {}   # stores the ASCII repr. of each DB
    matchingBuffer = {}   # keeps a pattern and result buffer for every query of each DB
    matchBuffer    = {}   # keeps a pattern and result buffer for every query of each DB
    matchOnly      = {}   # keeps the current list of possible match number (instead of using a special matchOnlyIn call to OBGrep)
    dbEntriesName    = {}   # stores the ASCII name (by comment) and the corresponding number
    dbEntriesNumber  = {}   # same as above, just vv
    
    

    def loadDB(self, fileName):
        
        dbFileName = fileName
        dbFile = open(dbFileName,'r')
        dbPlain = dbFile.readlines()
        dbFile.close()
        dbPlain = map(lambda x: x.strip(), dbPlain)
        dbNum = OBGrep.loadDB(dbFileName)
        self.dbCount += 1
        self.dbEntries[dbNum] = OBGrep.getNumEntries(dbNum)
        self.dbPlain[dbNum] = dbPlain
        self.matchingBuffer[dbNum] = {}
        self.matchBuffer[dbNum] = {}
        matchList = range(self.dbEntries[dbNum])
        matchList = map(lambda x: x+1, matchList)
        self.setCurrentMatchSet(dbNum, matchList)

        # save a dictionary with names of the compounds, enables getting compounds by name, rather than by number
        theNameMatch = re.compile('.*\s\#\s(\S+)')
        #names = map(lambda x: theNameMatch.search(x.strip()).groups()[0],dbPlain )
        names = None
        if names and len(names) == len(matchList):
            self.dbEntriesName[dbNum] = dict(map(None,names,matchList))
            self.dbEntriesNumber[dbNum] = dict(map(None,matchList,names))
            

        return dbNum

    def resetBuffers(self):
        for i in range(self.dbCount):
            self.matchingBuffer[i] = {}
            self.matchBuffer[i] = {}

    def getEntryComment(self, dbNum, entryNum):

        try:
            return self.dbEntriesNumber[dbNum][entryNum]
        except:
            return None

 
    def getEntryName(self, dbNum, entryNum):
        try:
            return self.dbEntriesNumber[dbNum][entryNum]
        except:
            return None

    def getEntryNumber(self, dbNum, entryName=''):
        try:
            return self.dbEntriesName[dbNum][entryName]
        except:
            return None
           
    def getEntryByName(self, dbNum, entryName=''):
        theEntryNumber = self.getEntryNumber(dbNum,entryName)
        if theEntryNumber:
            return self.getBestEntry(dbNum,theEntryNumber)

    def getEntryByNumber(self, dbNum, entryNumber=0):
        if entryNumber:
            return self.getBestEntry(dbNum,entryNumber)
        
    def getBestEntry(self, dbNum, entryNum):
        if self.dbEntries.has_key(dbNum) and entryNum > 0 and entryNum <=  self.dbEntries[dbNum]:
            entryString =  OBGrep.getEntryString(dbNum, entryNum).strip()

            # there seems to be a bug, as sometimes the returned string contains an [se] instead of [Se]
            # that might be the case if that is aromatic? Anyway, the SMARTS pattern containg [se] is not
            # valid. Therefore, we replace any [se] with [Se]

            entryString = entryString.replace('[se]','[Se]')

            # for similar reasons we are skipping [C@*@H?] and replace it by C

            
            entryString = entryString.replace('[C@@H]','C')
            entryString = entryString.replace('[C@H]','C')
            entryString = entryString.replace('[C@@]','C')
            entryString = entryString.replace('[C@]','C')

            # cut comments
            hash = entryString.find(' # ')
            if hash >= 0:
                entryString = entryString[:hash]


            return entryString

            #rest seems obsolete (ak:10.11.04)

            # in case it has disconnections (only works if the disconnection is not within a bracket)
            entryStringList = entryString.split('.')
            matching = []
            for entryS in entryStringList:
                try:
                    matching = OBGrep.getMatchingMode(dbNum,entryS,1,0)
                except Exception, e:
                    print "error e1 = %s\n" % e
                    print  entryNum, entryString, entryS
                    entryString = ''
                    pass
                
            if entryNum not in matching:
                #print '%s) not in results, use %s instead of %s ' % (entryNum, self.dbPlain[dbNum][entryNum-1], entryString)
                matching = []
                entryString = self.dbPlain[dbNum][entryNum-1]
                entryStringList = entryString.split('.')
                try:
                    for entryS in entryStringList:
                        matching = matching + OBGrep.getMatchingMode(dbNum,entryS,1,0)
                except Exception, e:
                    print "error e2 = %s\n" % e
                    print  'Cannot even take:', entryNum, entryString
                    entryString =''
                    pass
                if entryNum not in matching:
                    return entryString
                    raise  Exception, "The specified entry (num = %s) cannot be found using a SMARTS search" % entryNum
                
            return entryString
        return None

    def getNumEntries(self, dbNum, filter=1):
        if filter:
            return len(self.getCurrentMatchSet(dbNum))
        elif self.dbEntries.has_key(dbNum):
            return self.dbEntries[dbNum]

    def getMatching(self, dbNum, smarts, useFilter=1, useBuffer=0):
        stop = 0
        try:
            stop = smarts.index('mw')+1
        except:
            pass
        if stop:
            raise 'MW'
        if self.dbEntries.has_key(dbNum):
            matchList = []
            if useBuffer and self.matchingBuffer[dbNum].has_key(smarts):
                matchList =  self.matchingBuffer[dbNum][smarts]
            else:
                matches = OBGrep.getMatchingMode(dbNum,smarts,useFilter+1,0)
                self.matchingBuffer[dbNum][smarts] = matches
                matchList =  matches
            return matchList

            # currently not needed
            if useFilter and matchList:
                return self.filterMatchResult(dbNum, matchList)
                
            else:
                return matchList
            
        print "I do not have dbNum = %s as DB" % (dbNum)
        
            
    def match(self, dbNum, smarts, ):
        if self.dbEntries.has_key(dbNum):
            matchCount = []
            if self.matchBuffer[dbNum].has_key(smarts):
                matchCount =  self.matchBuffer[dbNum][smarts]
            else:
                count = OBGrep.match(dbNum,smarts)
                self.matchBuffer[dbNum][smarts] = count
                matchCount =  count
            return matchCount
            

    def setCurrentMatchSet(self, dbNum, matchSet):
        if matchSet is not None:
            self.matchOnly[dbNum] = matchSet
            OBGrep.setCurrentMatchSet(dbNum, matchSet)
            #self.resetBuffers()

    def getCurrentMatchSet(self, dbNum):
        if self.matchOnly.has_key(dbNum):
            return self.matchOnly[dbNum]+[]
        return None
            
    def subtractSet(self, dbNum, minusSet):
        currentMatchSet = self.getCurrentMatchSet(dbNum)
        if currentMatchSet:
            newMatchSet = filter(lambda x: x not in minusSet, currentMatchSet)
            self.setCurrentMatchSet(dbNum, newMatchSet)

    def filterMatchResult(self, dbNum, resultSet):
        currentMatchSet = self.getCurrentMatchSet(dbNum)
        if currentMatchSet:
            return  filter(lambda x: x in currentMatchSet, resultSet)
        else:
            return []
              
